package people;
import utility.ObscureWhichIsWhich;
import utility.One;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map.Entry;
import helper.DeterminationHelper;
/*	Copyright 2021 Ian Stuart

	This file is part of Three Industry ABCE Model.

	Three Industry ABCE Model is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	Three Industry ABCE Model is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with Three Industry ABCE Model.  If not, see <https://www.gnu.org/licenses/>.
*/

/**
 * The Person class is used to represent people in this Three Industry ABCE Model.
 * 
 * @author Ian Stuart
 */
public class Person extends AbstractAgent{
	private EReproductiveGender reproductiveGender;
	private boolean failedToMeetNeeds = false;
	private boolean alive = true;
	private Job job = null;	
	private boolean isAdult;
	private Home home= new Home();
	private HashMap<EBiologicalRelationship,BiologicalRelationship> biologicalRelationships = new HashMap<EBiologicalRelationship, BiologicalRelationship>();
	private RomanticRelationship romanticRelationship = null;
	private int money = 0;
	public static int totalFromAssistance = 0;
	public static int totalFromEmployerMinusIncomeTax = 0;
	public static int totalPaidInTaxes = 0;
	public static int totalPaidForNecessities = 0;
	public static int totalPaidForDiscretionaries = 0;
	public static int totalPaidForInheritanceTax = 0;
	
	public Person(EReproductiveGender reproductiveGender) {
		super(DeterminationHelper.getAdultAgeInMonths());
		this.reproductiveGender = reproductiveGender;
		this.money = DeterminationHelper.getAdultStartingMoney();
		this.isAdult = true;
	}
		
	public Person(EReproductiveGender reproductiveGender, int ageInMonths) {
		super(ageInMonths);
		this.reproductiveGender = reproductiveGender;
	}
	
	public void instantiate() {
		home.addPerson(this);
	}
	
	public boolean isAlive() {return this.alive;}
		public boolean isAdult() {
		if(! this.isAdult && this.getAge() >= DeterminationHelper.LEGAL_ADULT_AGE && this.alive) {
			this.isAdult = true;
			this.home.removePerson(this);
			this.home = new Home();
			this.home.addPerson(this);
		}
		return this.isAdult;
	}
	public Job getJob() {return this.job;}
	public void setJob(Job job) {this.job = job;}
	public int getMoney() {return this.money;}
	public void setMoney(int savings) {this.money = savings;}
	public void setHome(Home home) {this.home = home; if(this.home != null) this.home.addPerson(this);}
	public Home getHome() {return this.home;}
	public EReproductiveGender getReproductiveGender() {return this.reproductiveGender;}
	public void addBiologicalRelationship(EBiologicalRelationship eBiologicalRelationship, BiologicalRelationship biologicalRelationship) { this.biologicalRelationships.put(eBiologicalRelationship, biologicalRelationship);}
	public boolean hasBiologicalRelationship(EBiologicalRelationship biologicalRelationship) {
		return biologicalRelationships.containsKey(biologicalRelationship);
	}
	public BiologicalRelationship getBiologicalRelationship(EBiologicalRelationship biologicalRelationship) { return biologicalRelationships.get(biologicalRelationship);}
	public void setBiologicalRelationships(HashMap<EBiologicalRelationship,BiologicalRelationship> biologicalRelationships) { this.biologicalRelationships = biologicalRelationships; }
	public RomanticRelationship getRomanticRelationship() { return this.romanticRelationship; }
	public void setRomanticRelationship(RomanticRelationship romanticRelationship) { this.romanticRelationship = romanticRelationship;}
	
	/* 
	 * This method executes the main actions of the person agent. 
	 */
	public void executiveFunction() {
		doesGoToWork();
		satisfyNeeds();
		satisfyWants();
		seeIfPersonLives();
	}


	
	/*
	 * This method submits the person's monetary balance to the government.
	 */
	public void updateRecords() throws IllegalStateException {
		if(this.money < 0) {
			throw new IllegalStateException("A person cannot have a negative balance.");
		}
		THE_WORLD.getInstance().getGovernment().submitPersonalBalanceForRecord(this.money);
	}

	/* This method will see if there are jobs and if a person is accepted for employment they will have a job for the next time step.
	 * 
	 * This method priorities job openings in the Raw Material and Necessity industries. 
	 */
	public boolean lookForWork() {
		boolean success = false;
		Industry newEmployer = null;		
		//people have to prioritize working in the Raw Material Industry or the Necessity Industry for their own survival (regardless of pay);
		Industry rawMaterials = null;
		Industry necessities = null;
		if(THE_WORLD.getInstance().getIndustry(EConsumable.RAW_MATERIAL).getJobOpening() != null) {
			rawMaterials = THE_WORLD.getInstance().getIndustry(EConsumable.RAW_MATERIAL);				
		}
		if(THE_WORLD.getInstance().getIndustry(EConsumable.NECESSITY).getJobOpening() != null) {
			necessities = THE_WORLD.getInstance().getIndustry(EConsumable.NECESSITY);
		}
		//people have to prioritize working in the Raw Material Industry or the Necessity Industry for their own survival (regardless of pay);
		if(rawMaterials != null && necessities != null) {
			ObscureWhichIsWhich<Industry> obscureWhichIsWhich = new ObscureWhichIsWhich<Industry>(rawMaterials, necessities);
			newEmployer = obscureWhichIsWhich.getItemA();
		}else if(rawMaterials != null || necessities != null) {
			newEmployer = rawMaterials != null ? rawMaterials : necessities;
		}else if (THE_WORLD.getInstance().getIndustry(EConsumable.DISCRETIONARY).getJobOpening() != null) {
			newEmployer = THE_WORLD.getInstance().getIndustry(EConsumable.DISCRETIONARY);
		}
		if(newEmployer != null) {
			success = newEmployer.doesLikeAndAcceptApplication(this);
		}
		return success;
	}

	/* A person's employer will call this method in return for a person's work.
	 * The "pileOfMoney" argument should be the person's entire pay for a time unit.
	 * This method handles paying income tax. 
	 */
	public boolean doesAcceptPayment(One<Integer> pileOfMoney) {
		int originalAmount = pileOfMoney.get();
		boolean success = false;
		int amountTaxed = THE_WORLD.getInstance().getGovernment().acceptTaxOnIncomeAndReturnAmountOfTaxPaid(pileOfMoney);
		Person.totalFromEmployerMinusIncomeTax += pileOfMoney.get();
		Person.totalPaidInTaxes += amountTaxed;
		if(amountTaxed + pileOfMoney.get() != originalAmount || pileOfMoney.get() <= 0) {
			throw new IllegalStateException("Paying income tax descrepancy!");
		}
		if(pileOfMoney.get() > 0) {
			this.money = this.money + pileOfMoney.get();
			success = true;
		}
		return success;
	}

	public void acceptAssistance(One<Integer> pileOfMoney) {
		if(pileOfMoney.get() <= 0) {
			throw new IllegalArgumentException("A person cannot accept a zero or negative amount of money");
		}
		Person.totalFromAssistance += pileOfMoney.get();
		this.money = this.money + pileOfMoney.get();
	}

	public int getNumberOfKids() {
		int numberOfKids = 0;
		for(Entry<EBiologicalRelationship, BiologicalRelationship> biologicalRelationship : biologicalRelationships.entrySet()) {
			if(biologicalRelationship.getKey().isOffspring()) {
				numberOfKids++;
			}
		}
		return numberOfKids;
	}

	public ArrayList<Person> getLivingKids(){
		ArrayList<Person> kids = new ArrayList<Person>();
		for(Entry<EBiologicalRelationship, BiologicalRelationship> biologicalRelationship : biologicalRelationships.entrySet()) {
			if(biologicalRelationship.getKey().isOffspring() && biologicalRelationship.getValue().getTheOtherPerson(this).isAlive()) {
				kids.add(biologicalRelationship.getValue().getTheOtherPerson(this));
			}
		}
		return kids;
	}
	
	/*
	 * This method should be called each time step. This method determines if people die. If a person wasn't able to meet there needs they have a chance of dying. If a person has passed their life expectancy they die. 
	 */
	private void seeIfPersonLives() {
		if((this.failedToMeetNeeds && DeterminationHelper.doesDieFromFailingToMeetNeeds()) || this.getAge() > DeterminationHelper.LIFE_EXPECTANCY || !this.alive) {
			if(this.alive) {
				this.alive = false;
			}
			kill(this);
		}
	}
	
	/* This method handles weather a person goes to work, looks for work, or receives social security. If a person is working in the discretionary industry and there is a product/consumable shortage (and job opening)
	 * in the Raw Material or Necessity industry, the person will apply for and work at the industry with a shortage instead. This method will only return true if a person was paid for working at an industry.
	 * 
	 * If a person didn't have a job, didn't get paid, was retired, or collected social security this method returns false. 
	 */
	private boolean doesGoToWork() {
		if(job != null && this.getAge() >= THE_WORLD.getInstance().getGovernment().getRetirementAge()) {
			this.job.setPersonNull();
			this.job = null;
			return false;
		}else if(this.getAge() < THE_WORLD.getInstance().getGovernment().getRetirementAge()){
			if(job != null) {			
				if(!job.isEssential()){
					//people have to prioritize working in the Raw Material Industry or the Necessity Industry for their own survival (regardless of pay);
					boolean isRawMaterialSortageEmergencyAndThereIsAJobOpening = THE_WORLD.getInstance().getGovernment().isRawMaterialShortageEmergency() && 
																					THE_WORLD.getInstance().getIndustry(EConsumable.RAW_MATERIAL).getJobOpening() != null;
					boolean isNecessityShortageEmergencyAndThereIsAJobOpening = THE_WORLD.getInstance().getGovernment().isNecessityShortageEmergency() && 
																						THE_WORLD.getInstance().getIndustry(EConsumable.NECESSITY).getJobOpening() != null;
					if(isRawMaterialSortageEmergencyAndThereIsAJobOpening && isNecessityShortageEmergencyAndThereIsAJobOpening) {
						ObscureWhichIsWhich<Industry> obscureWhichIsWhich = new ObscureWhichIsWhich<Industry>(THE_WORLD.getInstance().getIndustry(EConsumable.RAW_MATERIAL), 
																													THE_WORLD.getInstance().getIndustry(EConsumable.NECESSITY));
						this.job.setPersonNull();
						this.job = null;
						obscureWhichIsWhich.getItemA();
					}else if(isRawMaterialSortageEmergencyAndThereIsAJobOpening) {
						this.job.setPersonNull();
						this.job = null;
						THE_WORLD.getInstance().getIndustry(EConsumable.RAW_MATERIAL).doesLikeAndAcceptApplication(this);
					}else if(isNecessityShortageEmergencyAndThereIsAJobOpening) {
						this.job.setPersonNull();
						this.job = null;
						THE_WORLD.getInstance().getIndustry(EConsumable.NECESSITY).doesLikeAndAcceptApplication(this);
					}
				}
				
				if(job != null && job.getIndustry().doesAcceptExpectedWork(this)) {
					return true;
				}
			}else {
				lookForWork();
			}
			return false;
		}else {
			THE_WORLD.getInstance().getGovernment().doesPaySocialSecurity(this);
			return false;
		}
	}
	
	/*
	 * This method handles satisfying the person's needs for the time step. The required necessities for everyone in a "home" is paid for only once in a time step. If this is successful the method returns true.
	 * If the required necessity consumable was not paid for this method returns false. 
	 */
	private boolean satisfyNeeds() {
		boolean success = false;
		int amountNeeded = 0;
		int moneyNeeded = 0;
		int transferAmount = 0;
		Person adult = this;
		if(!adult.getHome().didPayForNeeds() && THE_WORLD.getInstance().getTime() != 0) {
			if(!adult.isAdult()) {
				BiologicalRelationship rel = adult.getBiologicalRelationship(EBiologicalRelationship.MOTHER);
				if(rel == null || !rel.getTheOtherPerson(this).isAlive()) {
					rel = adult.getBiologicalRelationship(EBiologicalRelationship.FATHER);
					if(rel != null && rel.getTheOtherPerson(this).isAlive()) {
						adult = rel.getTheOtherPerson(adult);
					}
				}else if(rel.getTheOtherPerson(this).isAlive()){
					adult = rel.getTheOtherPerson(adult);
				}else {
					adult = null;
				}
			}
			if(adult != null) {
				amountNeeded = adult.getHome().getRequiredAmountNecessityForATimeUnitOfShelter(adult);
				if(THE_WORLD.getInstance().getIndustry(EConsumable.NECESSITY).hasQuantity(amountNeeded)) {
					int moneyExcludingSalesTax = (int)THE_WORLD.getInstance().getIndustry(EConsumable.NECESSITY).getPriceForConsumableExcludingSalesTax(EConsumable.NECESSITY, amountNeeded);
					moneyNeeded = THE_WORLD.getInstance().getIndustry(EConsumable.NECESSITY).getTotalPriceForConsumable(EConsumable.NECESSITY, amountNeeded);
					if(moneyNeeded > adult.getMoney() && adult.getRomanticRelationship() != null && moneyNeeded < adult.getMoney() + adult.getRomanticRelationship().getTheOtherPerson(adult).getMoney()) {
						/*int*/ transferAmount = moneyNeeded - adult.getMoney();
						adult.getRomanticRelationship().getTheOtherPerson(adult).setMoney(adult.getRomanticRelationship().getTheOtherPerson(adult).getMoney() - transferAmount);
						adult.setMoney(adult.getMoney() + transferAmount);
					}else if(moneyNeeded > adult.getMoney()) {
						int amountToRequest = moneyNeeded - adult.getMoney();
						boolean yes = THE_WORLD.getInstance().getGovernment().doesAcceptAndProvideAssistanceUponRequest(adult, amountToRequest);
						if(yes) {
							Person.totalFromAssistance += amountToRequest;
						}
					}
					if(moneyNeeded <= adult.getMoney()) {
						One<Integer> pileOfMoney = new One<Integer>(moneyNeeded);
						if(THE_WORLD.getInstance().getIndustry(EConsumable.NECESSITY).doesAcceptPurchaseOfConsumable(EConsumable.NECESSITY, amountNeeded, pileOfMoney)){
							Person.totalPaidForNecessities += moneyExcludingSalesTax;
							Person.totalPaidInTaxes += moneyNeeded - moneyExcludingSalesTax;
							adult.setMoney(adult.getMoney() - moneyNeeded);
							success = true;
							adult.getHome().setNeedsPaidFor();
						}
					}else if(transferAmount > 0) {
						System.out.println("transferError");
					}
				}
			}
		}else {
			success = true;
		}
		if(!success) {
			this.failedToMeetNeeds = true;
		}
		return success;
	}
	
	/*
	 * If the person has enough extra money this method will purchase as many of the discretionary consumable as they can afford. 
	 */
	private boolean satisfyWants() {
		boolean success = false;	
		double determineMoneyNeeded = THE_WORLD.getInstance().getIndustry(EConsumable.DISCRETIONARY).getPriceForConsumableExcludingSalesTax(EConsumable.DISCRETIONARY, 1);
		int amountCanAfford = 0;
		if(determineMoneyNeeded == 0) {
			amountCanAfford = DeterminationHelper.AMOUNT_OF_DISCRETIONARY_A_PERSON_GETS_IF_FREE;
		}else{
			determineMoneyNeeded = determineMoneyNeeded + (determineMoneyNeeded * THE_WORLD.getInstance().getGovernment().getSalesTaxRate());
			amountCanAfford = (int)(((double)(this.getMoney() - DeterminationHelper.AMOUNT_OF_MONEY_PEOPLE_MAKE_SURE_TO_SAVE_WHEN_SPENDING_MONEY_ON_DESCRETIONARIES))/determineMoneyNeeded);
			amountCanAfford /= 3;
		}
		if(amountCanAfford > 0 && THE_WORLD.getInstance().getIndustry(EConsumable.DISCRETIONARY).hasQuantity(amountCanAfford)) {
			int moneyNeeded = THE_WORLD.getInstance().getIndustry(EConsumable.DISCRETIONARY).getTotalPriceForConsumable(EConsumable.DISCRETIONARY, amountCanAfford);
			int priceForOnlyItems = (int)THE_WORLD.getInstance().getIndustry(EConsumable.DISCRETIONARY).getPriceForConsumableExcludingSalesTax(EConsumable.DISCRETIONARY, amountCanAfford);
			One<Integer> pileOfMoney = new One<Integer>(moneyNeeded);
			if(THE_WORLD.getInstance().getIndustry(EConsumable.DISCRETIONARY).doesAcceptPurchaseOfConsumable(EConsumable.DISCRETIONARY, amountCanAfford, pileOfMoney)){
				Person.totalPaidForDiscretionaries += priceForOnlyItems;
				Person.totalPaidInTaxes += moneyNeeded - priceForOnlyItems;
				this.setMoney( this.getMoney() - moneyNeeded);
				success = true;
			}
		}
		return success;
	}
	
	/*
	 * The person should pass themselves in as the "person" argument. If the person is not alive this method will handle inheritance and taxes and remove them from the list of living people. 
	 */
	private void kill(Person person) {
		if(!person.isAlive() && person == this) {
			One<Integer> inheritance = new One<Integer>(person.getMoney());
			int originalAmount = inheritance.get();
			int inheritanceGiven = 0;
			int taxPaid = 0;
			if(inheritance.get() > 0) {
				ArrayList<Person> kids = person.getLivingKids();
				if(kids.size() > 0) {
					taxPaid = THE_WORLD.getInstance().getGovernment().payInheritanceTaxAndReturnAmountOfTaxPaid(inheritance);
					Person.totalPaidForInheritanceTax += taxPaid;
					Person.totalPaidInTaxes += taxPaid;				
					int portion = inheritance.get()/kids.size();
					for(int i = 0; i < kids.size(); i++) {
						if(i == kids.size() - 1) {
							kids.get(i).setMoney(kids.get(i).getMoney() + inheritance.get());
							inheritanceGiven += inheritance.get();
						}else {
							kids.get(i).setMoney(kids.get(i).getMoney() + portion);
							inheritance.set(inheritance.get() - portion);
							inheritanceGiven += portion;
						}
					}
				}else {
					taxPaid = THE_WORLD.getInstance().getGovernment().pay100PercentInheritanceTaxBecauseYouDidntHaveLivingKidsAndReturnAmountOfTaxPaid(inheritance);
					Person.totalPaidForInheritanceTax += taxPaid;
				}
			}
			if(inheritanceGiven + taxPaid != originalAmount) {
				throw new IllegalStateException("Paying inheritance tax descrepancy!");
			}
			person.setMoney(0);
			THE_WORLD.getInstance().removePerson(person);
		}
	}
	
	/*
	 * The Home class ties people together that shelter together (people in romantic relationships and their children) and determines how much of the Necessity consumable is needed for the people of one home in a time unit. 
	 */
	public class Home {
		public ArrayList<Person> people = new ArrayList<Person>();
		int lastTimeUnitNeedsWerePaidFor = -1;
		
		public Home() {}
		
		public void setNeedsPaidFor() { lastTimeUnitNeedsWerePaidFor = THE_WORLD.getInstance().getTime(); }
		public boolean didPayForNeeds() { return lastTimeUnitNeedsWerePaidFor == THE_WORLD.getInstance().getTime(); }
		
		public int getRequiredAmountNecessityForATimeUnitOfShelter(Person person) {
			double requiredAmountOfNecessityConsumable = DeterminationHelper.REQUIRED_NECESSITY_CONSUMABLE_FOR_FIRST_ADULT_OF_A_HOME;
			for(int i = 1; i < people.size(); i++) {
				if(people.get(i).isAdult()){
					requiredAmountOfNecessityConsumable += DeterminationHelper.ADDITIONAL_NECESSITY_CONSUMABLE_REQUIRED_FOR__THE_SECOND_ADULT_OF_A_HOME_WHERE_THE_TOTAL_IS_ROUNDED_DOWN;
				}else {
					requiredAmountOfNecessityConsumable += DeterminationHelper.ADDITIONAL_NECESSITY_CONSUMABLE_REQUIRED_FOR_EACH_PERSON_UNDER_18_WHERE_THE_TOTAL_FOR_A_HOME_IS_ROUNDED_DOWN;
				}
			}
			return (int)requiredAmountOfNecessityConsumable;
		}
		
		public void addPerson(Person person) {
			if(!people.contains(person)) {
				people.add(person);
				person.setHome(this);
			}
		}
		
		public void removePerson(Person person) {
			people.remove(person);
		}
		
	}
}